use Package::Alias 'Config::IOD' => 'Config::INI::OnDrugs';
package Config::IOD; # to make PodWeaver happy
1;
#ABSTRACT: Alias for Config::Ini::OnDrugs

__END__
=pod

=head1 NAME

Config::IOD - Alias for Config::Ini::OnDrugs

=head1 VERSION

version 0.02

=head1 AUTHOR

Steven Haryanto <stevenharyanto@gmail.com>

=head1 COPYRIGHT AND LICENSE

This software is copyright (c) 2011 by Steven Haryanto.

This is free software; you can redistribute it and/or modify it under
the same terms as the Perl 5 programming language system itself.

=cut

